#!/usr/bin/python
# Script for local logon to Ezlo hub via Websockets. User and Token are stored in file for reuse between sessions.
# Written by Rene Boer , 10 May 202
# Version 1.1
#
# Known issues:
# where to get the PK_Account for remote access. It is not in the documented calls ??
# Workaround is to skip getting list of controlelrs on account and assuming Server_Account and Server_Device are the same (they are for me). 
# Use option --help for full usage details.

debug = False
certverify = False

import ssl
import websocket
from websocket import create_connection
import requests
import json
import sys
import hashlib
import urllib3
import uuid
import os
import http.client
import logging
import argparse
import base64

# WS port to use for local access
ezlo_port = '17000'

# Storage for local access credentials between sessions.
ezlo_keys_store = ".ezlo_local_keys"

# login to the portal and establish WebSocket connection via the relay server.
def EzloLoginRemote(user_id, password, controller_name):
    Ezlo_MMS_salt = "oZ7QE6LcLJp6fiWzdqZc"
    authentication_url = 'https://vera-us-oem-autha11.mios.com/autha/auth/username/{user}?SHA1Password={pwd}&SHA1PasswordCS={pwdCS}&PK_Oem=1&TokenVersion=2'
    get_controllers_url = 'https://{server}/account/account/account/{account}/devices'
    get_device_url = 'https://{server}/device/device/device/{device}'

    # The controller name can have PK_Account as postfix as I do not know where to get it from. Not needed for V1.1
    #if len(controller_name.split(',')) == 2:
    #    pk_account = controller_name.split(',')[1]
    #    controller_name = controller_name.split(',')[0]
    # Get Tokens
    request_headers = {
        'Accept': 'application/json',
        'Content-Type': 'application/json;charset=UTF-8',
        'Access-Control-Allow-Origin': '*',
        'User-Agent': 'RB Vera Bridge 1.0'
    }
    hash = user_id.lower()+password+Ezlo_MMS_salt
    SHA1pwd = hashlib.sha1(hash.encode())
    hash = user_id+password+Ezlo_MMS_salt
    SHA1pwdCS = hashlib.sha1(hash.encode())
    print("====================================================================")
    print("Logon to portal for MMS Keys")
    #print(authentication_url.format(user=user_id,pwd=SHA1pwd.hexdigest()))
    response = requests.get(authentication_url.format(user=user_id,pwd=SHA1pwd.hexdigest(),pwdCS=SHA1pwdCS.hexdigest()), headers=request_headers)
    if response.status_code != 200:
        return 'cannot logon to portal',0
    if debug:
        print("\n")
        print(response.text)
        print("====================================================================")
    js_resp = response.json()
    MMSAuth = js_resp.get('Identity')
    MMSAuthSig = js_resp.get('IdentitySignature')
    server_acc = js_resp.get('Server_Account')
    request_headers["MMSAuth"] = MMSAuth
    request_headers["MMSAuthSig"] = MMSAuthSig
	# Identity is base64 encoded json with account data
	# Identity remains valid for 24 days it seems.
    # Identity contains:	#{"Expires":1589635592,"Generated":1589549192,"PermissionsEnabled":[1,2,3.......626,1648,{"PK":1653,"Arguments":"[9,10,11,12,13]"}],"PermissionsDisabled":[],"Version":2,"PK_Account":6xxx,"PK_AccountType":5,"PK_AccountChild":0,"PK_Account_Parent":2375,"PK_Account_Parent2":1,"PK_Account_Parent3":0,"PK_Account_Parent4":0,"PK_App":0,"PK_Oem":1,"PK_Oem_User":"NULL","PK_PermissionRole":10,"PK_User":1xxx,"PK_Server_Auth":1,"PK_Server_Account":6,"PK_Server_Event":14,"Server_Auth":"vera-us-oem-autha11.mios.com","Seq":1xxxxx,"Username":"rxxxxx"}
    identity = base64.b64decode(MMSAuth)
    print(identity)
    js_account = json.loads(identity.decode("utf-8"))
    pk_account = js_account.get('PK_Account')
    print("\n====================================================================")
    print("Get Controllers List")
    response = requests.get(get_controllers_url.format(server=server_acc, account=pk_account), headers=request_headers)
    if response.status_code != 200:
        return 'cannot get list of controllers',0
    if debug:
        print("\n")
        print(response.text)
        print("====================================================================\n")
    js_resp = response.json()
    server_acc = ''
    for device in js_resp.get('Devices'):
        if device.get('PK_Device') == controller_name:
            server_acc = device.get('Server_Device')
            controller_id = device.get('PK_Device')

    # Make sure we found the controller
    if server_acc == '':
        return 'controller with serial not found '+controller_name,0
    # Get controllers details
    print("\n====================================================================")
    print("Get Controller Details")
    response = requests.get(get_device_url.format(server=server_acc, device=controller_id), headers=request_headers)
    if response.status_code != 200:
        return 'cannot get controller details',0
    if debug:
        print("\n")
        print(response.text)
        print("====================================================================\n")
    js_resp = response.json()
    server_relay = js_resp.get('Server_Relay')
    if js_resp.get('LinuxFirmware') != 1:
        return 'not a linux controller',0

	# Connect to web sockets
    print("\n====================================================================")
    print("Connect to web sockets")
    ws = create_connection(server_relay, sslopt={"cert_reqs": ssl.CERT_NONE, "check_hostname": False})
    cmd= '{"method": "loginUserMios", "id": "loginUserMios_12","params": {"MMSAuth":"'+MMSAuth+'","MMSAuthSig":"'+MMSAuthSig+'","PK_Device": "'+controller_id+'"}}'
    if debug:
        print("\n")
        print(cmd)
        print("\n")
    ws.send(cmd)
    resp = ws.recv()
    js_resp = json.loads(resp)
    if js_resp.get('error'):
        return 'cannot login to Hub '+js_resp.get('error'),0
    if debug:
        print(js_resp)
        print("====================================================================\n")
		
	# Register to device
    print("\n====================================================================")
    print("Register to device")
    cmd= '{"method": "register", "id": "register_1234", "params": {"serial":"'+controller_id+'"}}'
    if debug:
        print("\n")
        print(cmd)
        print("\n")
    ws.send(cmd)
    resp = ws.recv()
    js_resp = json.loads(resp)
    if js_resp.get('error'):
        return 'cannot register to Hub '+js_resp.get('error'),0
    if debug:
        print(js_resp)
    print("====================================================================\n")
    print("WebSocket Connection open.\n")
    return 'OK', ws

def EzloLoginLocal(user_id, password, controller_ip, serial):
    # If we have the logon data for local, no need to get it again.
    wss_user = ''
    wss_token = ''
    # See if we have user and token key stored
    if os.path.isfile(ezlo_keys_store):
        fp = open(ezlo_keys_store)
        ts = fp.read()
        fp.close()
        js_ts = json.loads(ts)
        wss_user = js_ts.get('user')
        wss_token = js_ts.get('token')

    # Not stored, logon to portal to get them.
    if wss_user == '' or wss_token == '':
        Ezlo_MMS_salt = "oZ7QE6LcLJp6fiWzdqZc"
        authentication_url = 'https://vera-us-oem-autha11.mios.com/autha/auth/username/{user}?SHA1Password={pwd}&SHA1PasswordCS={pwdCS}&PK_Oem=1&TokenVersion=2'
        get_token_url = 'https://cloud.ezlo.com/mca-router/token/exchange/legacy-to-cloud/'
        sync_token_url = 'https://api-cloud.ezlo.com/v1/request'

        # Get Tokens
        request_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json;charset=UTF-8',
            'Access-Control-Allow-Origin': '*',
            'User-Agent': 'RB Vera Bridge 1.0'
        }
        hash = user_id.lower()+password+Ezlo_MMS_salt
        SHA1pwd = hashlib.sha1(hash.encode())
        hash = user_id+password+Ezlo_MMS_salt
        SHA1pwdCS = hashlib.sha1(hash.encode())
        print("====================================================================")
        print("Logon to portal for MMS Keys")
        #print(authentication_url.format(user=user_id,pwd=SHA1pwd.hexdigest()))
        response = requests.get(authentication_url.format(user=user_id,pwd=SHA1pwd.hexdigest(),pwdCS=SHA1pwdCS.hexdigest()), headers=request_headers)
        if response.status_code != 200:
            return 'cannot logon to portal',0
        if debug:
            print("\n")
            print(response.text)
            print("====================================================================\n")
        js_resp = response.json()

        print("====================================================================")
        print("Get Token")
        MMSAuth = js_resp.get('Identity')
        MMSAuthSig = js_resp.get('IdentitySignature')
        request_headers["MMSAuth"] = MMSAuth
        request_headers["MMSAuthSig"] = MMSAuthSig
        response = requests.get(get_token_url, headers=request_headers, verify=certverify)
        if response.status_code != 200:
            return 'cannot get token',0
        if debug:
            print("\n")
            print(response.text)
            print("====================================================================\n")
        js_resp = response.json()
        token=js_resp.get('token')
	
        # Get controller keys (user & token)
        print("====================================================================")
        print("Get controller keys")
        new_uuid = uuid.uuid4()
        post_headers = {
            'Authorization': 'Bearer '+token,
            'Accept': 'application/json',
            'Content-Type': 'application/json;charset=UTF-8',
            'Access-Control-Allow-Origin': '*',
            'User-Agent': 'RB Vera Bridge 1.0'
        }
        post_data = {
            'call': 'access_keys_sync',
            'version': '1',
            'params': {
                'version': 53, 
                'entity': 'controller',
                'uuid': str(new_uuid)
            }
        }
        response = requests.post(sync_token_url, json=post_data, headers=post_headers, verify=certverify)
        if response.status_code != 200:
            return '',0
        if debug:
            print("\n")
            print(response.text)
            print("====================================================================\n")
        js_resp = response.json()
        # Get user and token from response.
        data = js_resp.get('data')
		# first look up controller uuid
        contr_uuid = ''
        for key in data.get('keys'):
            key_data = data.get('keys').get(key)
            if key_data.get('meta'):
                if key_data.get('meta').get('entity'):
                    if key_data.get('meta').get('entity').get('id'):
                        if key_data.get('meta').get('entity').get('id') == serial:
                            contr_uuid = key_data.get('meta').get('entity').get('uuid')
        if contr_uuid == '':
            print("Controller serial not found\n")
            return '',0
        for key in data.get('keys'):
            key_data = data.get('keys').get(key)
            if key_data.get('data'):
                if key_data.get('data').get('string'):
                    if key_data.get('meta').get('target').get('uuid') == contr_uuid:
                        wss_token = key_data.get('data').get('string')
                        wss_user = key_data.get('meta').get('entity').get('uuid')
                        wss_data = wss_user+":"+wss_token
                        wss_authorization = str(base64.b64encode(wss_data.encode("utf-8")), "utf-8")
                        # Write keys to file for reuse
                        fp = open(ezlo_keys_store, "w")
                        fp.write('{"user": "'+wss_user+'","token":"'+wss_token+'"}')
                        fp.close()
                        break
    else:
	    print("Using credentials from file.")

	# Open web socket
    print("====================================================================")
    print("Open web socket")
    ws = create_connection('wss://' + controller_ip + ':' + ezlo_port, sslopt={"cert_reqs": ssl.CERT_NONE, "check_hostname": False})

	# Connect to controller using the keys
    print("====================================================================")
    print("Connect to controller using the keys")
    cmd= '{"method": "hub.offline.login.ui", "id": "ID0", "params": {"user":"'+wss_user+'","token":"'+wss_token+'"}}'
    if debug:
        print("\n")
        print(cmd)
        print("\n")
    ws.send(cmd)
    resp = ws.recv()
    js_resp = json.loads(resp)
    if debug:
        print("\n")
        print(js_resp)
    # {"error":null,"id":"ID0","method":"hub.offline.login.ui","result":{}}
    # {"error":{"code":-32500,"data":"user.login.badpassword","message":"Bad password"},"id":"ID0","method":"hub.offline.login.ui"}
    if js_resp.get('error'):
        if os.path.isfile(ezlo_keys_store):
            # Remove stored key file as they are wrong.
            os.remove(ezlo_keys_store)
        print(js_resp.get('error'))	
        return '',0
    print("====================================================================\n")
    print("WebSocket Connection open.\n")
    return 'OK', ws

if __name__ == '__main__':
    # parse arguments
    parser = argparse.ArgumentParser(description='Connect to Ezlo Hub using WebSockets')
    group = parser.add_mutually_exclusive_group()
    parser.add_argument('serial', help='Your Ezlo Hub serial number.')
    parser.add_argument('user', help='Your Ezlo Hub user id.')
    parser.add_argument('password', help='Your Ezlo Hub password.')
    group.add_argument('-l', '--local', metavar='IP Address', help='Connect locally to IP Address.')
    parser.add_argument('-d', '--debug', action="store_true", help='Show debug commands.')
    args = parser.parse_args()
    user_id = args.user
    pwd = args.password
    controller_ip = args.local
    controller_serial = args.serial
    if args.debug:
        debug = True

    # Enable debugging of http requests (gives more details on Python 2 than 3 it seems)    
    if debug:
        http.client.HTTPConnection.debuglevel = 1
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True
        websocket.enableTrace(True)

    msg = 'No Hub serial or IP address specified'
    if controller_ip:
        msg, ws = EzloLoginLocal(user_id, pwd, controller_ip, controller_serial)
    if controller_serial:
        msg, ws = EzloLoginRemote(user_id, pwd, controller_serial)
    if msg != 'OK':
        print('Failed to login ', msg)
        sys.exit()

    # List the devices
    # Send list devices command
    print("Send list devices")
    cmd= '{ "method": "hub.devices.list", "id": "ID1" , "params": {} }'
    if debug:
        print("\n")
        print(cmd)
        print("\n")
    ws.send(cmd)
    print("\n")
    print(ws.recv())
    print("\n")

    # Send list items command
    print("Send list items")
    cmd= '{ "method": "hub.items.list", "id": "ID2" , "params": {} }'
    if debug:
        print("\n")
        print(cmd)
        print("\n")
    ws.send(cmd)
    print("\n")
    print(ws.recv())
    print("\n")
		
